from time import clock
import warnings

from itertools import repeat
import concurrent.futures
import multiprocessing

import click
import numpy as np
import rasterio
from crick import TDigest
from scipy import stats
from scipy.stats.mstats import mquantiles
import matplotlib.pyplot as plt

import rio_terrain as rt
import rio_terrain.tools.messages as msg
from rio_terrain import __version__ as terrain_version


def tdigest_mean(digest):
    """Estimate the mean from a tdigest

    """
    ctr = digest.centroids()
    mean = np.sum(ctr['mean']*ctr['weight'])/np.sum(ctr['weight'])

    return mean


def tdigest_std(digest):
    """Estimate the standard deviation of the mean from a tdigest

    """
    ctr = digest.centroids()
    mean = tdigest_mean(digest)
    std = np.sqrt(
                (np.power((ctr['mean'] - mean), 2)*ctr['weight']).sum()
                / (ctr['weight'].sum() - 1)
                )

    return std


def tdigest_stats(digest):
    """Estimate descriptive statistics from a tdigest

    """
    min = digest.min()
    max = digest.max()
    mean = tdigest_mean(digest)
    std = tdigest_std(digest)

    return (min, max, mean, std)


def digest_window(file, window, absolute):
    '''Process worker that calculates a t-digest on a raster

    '''
    with rasterio.open(file) as src:
        data = src.read(1, window=window[1])
        data[data <= src.nodata+1] = np.nan
        arr = data[np.isfinite(data)]
        if absolute:
            arr = np.absolute(arr)
        count_ = np.count_nonzero(~np.isnan(data))
        digest_ = TDigest()
        digest_.update(arr.flatten())

    return window, digest_, count_


@click.command()
@click.argument('input', nargs=1, type=click.Path(exists=True))
@click.option('-q', '--quantile', multiple=True, type=float,
              help='Print quantile value')
@click.option('-f', '--fraction', nargs=1, default=1.0,
              help='Randomly sample a fraction of data blocks')
@click.option('--absolute/--no-absolute', default=False,
              help='Calculate quantiles based on the set of absolute values')
@click.option('--describe/--no-describe', default=False,
              help='Print descriptive statistics to the console')
@click.option('--plot/--no-plot', default=False,
              help='Display statistics plots')
@click.option('-j', '--jobs', 'njobs', type=int, default=multiprocessing.cpu_count(),
              help='Number of concurrent jobs to run')
@click.option('-v', '--verbose', is_flag=True, help='Enables verbose mode')
@click.version_option(version=terrain_version, message='%(version)s')
@click.pass_context
def quantiles(ctx, input, quantile, fraction, absolute, describe, plot, njobs, verbose):
    """Calculates and prints quantile values.

    \b
    Example:
    rio quantiles elevation.tif -q 0.5 -q 0.9

    """

    t0 = clock()

    count = 0

    with rasterio.Env():

        with rasterio.open(input) as src:
            profile = src.profile
            affine = src.transform
            step = (affine[0], affine[4])

            if njobs < 1:

                click.echo('Computing quantiles using scipy...')
                data = src.read(1)
                data[data <= src.nodata+1] = np.nan
                arr = data[np.isfinite(data)]
                if absolute:
                    arr = np.absolute(arr)

                count = np.count_nonzero(~np.isnan(data))
                description = (arr.min(), arr.max(), arr.mean(), arr.std())
                results = zip(quantile, mquantiles(arr, np.array(quantile)))

            elif njobs == 1:

                blocks = rt.subsample(src.block_windows(), probability=fraction)
                digest = TDigest()

                click.echo('Computing quantiles using t-digest...')
                for ij, window in blocks:
                    data = src.read(1, window=window)
                    data[data <= src.nodata+1] = np.nan
                    arr = data[np.isfinite(data[:])]
                    if absolute:
                        arr = np.absolute(arr)

                    window_count = np.count_nonzero(~np.isnan(data))
                    if window_count > 0:
                        window_digest = TDigest()
                        window_digest.update(arr.flatten())
                        digest.merge(window_digest)
                        count += window_count

                description = tdigest_stats(digest)
                results = zip(quantile, digest.quantile(quantile))

            else:

                blocks = rt.subsample(src.block_windows(), probability=fraction)
                digest = TDigest()

                click.echo('Computing quantiles using multiprocess t-digest...')
                with concurrent.futures.ProcessPoolExecutor(max_workers=njobs) as executor:
                    for (window, window_digest, window_count) in executor.map(digest_window, repeat(input), blocks, repeat(absolute)):
                        if window_count > 0:
                            digest.merge(window_digest)
                            count += window_count

                description = tdigest_stats(digest)
                results = zip(quantile, digest.quantile(quantile))

    t1 = clock()
    click.echo('Finished in : {}'.format(msg.printtime(t0, t1)))

    click.echo(list(results))
    min, max, mean, std = description

    if describe:
        click.echo('min:', min)
        click.echo('max:', max)
        click.echo('mean:', mean)
        click.echo('std:', std)
        click.echo('count:', count)

    if njobs > 0 and plot is True:

        ctr = digest.centroids()

        # scaled to theoretic normal
        # calculate positions relative to standard normal distribution
        qx_predicted_norm = stats.norm.ppf(digest.cdf(ctr['mean']))
        qx_norm = np.linspace(start=stats.norm.ppf(0.001), stop=stats.norm.ppf(0.999), num=250)
        qz_norm = qx_norm*std + mean
        cum_norm = stats.norm.cdf(qx_norm)

        # scaled to theoretic laplace
        # calculate positions relative to laplace distribution
        """
        qx_predicted_laplace = stats.laplace.ppf(digest.cdf(ctr['mean']))
        qx_laplace = np.linspace(start=stats.laplace.ppf(0.001), stop=stats.laplace.ppf(0.999), num=250)
        qz_laplace = qx_laplace*std + mean
        cum_laplace = stats.laplace.cdf(qx_laplace)
        """

        # frequency at centroid (irregular width bins)
        plt.plot(ctr['mean'], ctr['weight'], 'r.')
        plt.plot([mean], [0], 'k+', markersize=12)
        plt.xlabel('Centroid Value')
        plt.ylabel('Counts')
        plt.title('Centroid Counts')
        plt.show()

        # histogram (equal width bins)
        nbins = 1000
        hist, bin_edges = digest.histogram(nbins)
        width = (digest.max() - digest.min())/nbins
        plt.bar(bin_edges[:-1], hist, width=width)
        plt.xlabel('Value')
        plt.ylabel('Counts')
        plt.title('Histogram')
        plt.show()

        # cumulative probability distribution
        samples = np.arange(ctr['mean'].min(), ctr['mean'].max(), 1.0)
        cdf = digest.cdf(samples)

        plt.plot(samples, cdf, 'r.')
        plt.plot(qz_norm, cum_norm, linestyle='dashed', c='black')
        # plt.plot(qz_laplace, cum_laplace, linestyle='dotted', c='gray')
        plt.plot([mean], [digest.cdf(mean)], 'k+', markersize=12)
        plt.xlabel('Value')
        plt.ylabel('Probability')
        plt.title('Cumulative Distribution')
        plt.show()

        # theoretic normal
        plt.plot(qx_predicted_norm, ctr['mean'], 'r.')
        plt.plot(qx_norm, qz_norm, linestyle='dashed', c='black')
        plt.plot([0], [mean], 'k+', markersize=12)
        plt.xlabel('Standard Normal Variate')
        plt.ylabel('Value')
        plt.title('QQ-plot on theoretic standard normal')
        plt.show()

        # theoretic laplace
        """
        plt.plot(qx_predicted_laplace, ctr['mean'], 'r.')
        plt.plot(qx_laplace, qz_laplace, linestyle='dashed', c='black')
        plt.plot([0], [mean], 'k+', markersize=12)
        plt.xlabel('laplace variate')
        plt.ylabel('Elevation (m)')
        plt.title('QQ-plot on theoretic laplace')
        plt.show()
        """
    if njobs < 1 and plot is True:

        # histogram (equal width bins)
        nbins = 100
        hist, bin_edges = np.histogram(arr, nbins)
        width = (arr.max() - arr.min())/nbins
        plt.bar(bin_edges[:-1], hist, width=width)
        plt.xlabel('Elevation (m)')
        plt.ylabel('Counts')
        plt.title('Histogram')
        plt.show()

        # cumulative probability distribution
        cdf = np.cumsum(hist)/count
        qx_predicted_norm = stats.norm.ppf(cdf)

        qx_norm = np.linspace(start=stats.norm.ppf(0.001), stop=stats.norm.ppf(0.999), num=250)
        qz_norm = qx_norm*std + mean
        cum_norm = stats.norm.cdf(qx_norm)

        plt.plot(bin_edges[:-1], cdf, 'r.')
        plt.plot(qz_norm, cum_norm, linestyle='dashed', c='black')
        plt.plot([mean], [0], 'k+', markersize=12)
        plt.xlabel('Value')
        plt.ylabel('Probability')
        plt.title('Cumulative Distribution')
        plt.show()

        # theoretic normal

        # full dataset is too large!
        # zscore = (arr - mean)/std
        # plt.plot(zscore, arr, 'r.')

        plt.plot(qx_predicted_norm, bin_edges[:-1], 'r.')
        plt.plot(qx_norm, qz_norm, linestyle='dashed', c='black')
        plt.plot([0], [mean], 'k+', markersize=12)
        plt.xlabel('Standard Normal Variate')
        plt.ylabel('Value')
        #plt.yscale('symlog')
        plt.title('QQ-plot on theoretic standard normal')
        plt.show()



from time import clock
import warnings
import concurrent.futures
import multiprocessing

import click
import numpy as np
import rasterio

import rio_terrain as rt
import rio_terrain.tools.messages as msg
from rio_terrain import __version__ as terrain_version


def _slice(data, minimum=None, maximum=None, keep_data=False, false_val=0):
    """Slice data or ones from an array given a value range.

    Parameters:
        data (ndarray)
        minimum (float)
        maximum (float)
        keep_data (bool)

    Returns:
        result (ndarray)

    """

    # default bounds
    if minimum is None:
        minimum = np.nanmin(data)
    if maximum is None:
        maximum = np.nanmax(data)

    if keep_data:
        result = np.where((data >= minimum) & (data <= maximum), data, false_val)
    else:
        result = np.where((data >= minimum) & (data <= maximum), 1, false_val)

    return result


@click.command()
@click.argument('input', nargs=1, type=click.Path(exists=True))
@click.argument('output', nargs=1, type=click.Path())
@click.option('--minimum', nargs=1, type=float, default=None,
              help='Minimum value to extract.')
@click.option('--maximum', nargs=1, type=float, default=None,
              help='Maximum value to extract.')
@click.option('--keep-data', is_flag=True,
              help='Return the input data. Default is to return ones.')
@click.option('--zeros/--no-zeros', is_flag=True,
              help='Use the raster nodata value or zeros for False condition')
@click.option('-j', '--njobs', type=int, default=multiprocessing.cpu_count(),
              help='Number of concurrent jobs to run')
@click.option('-v', '--verbose', is_flag=True, help='Enables verbose mode.')
@click.version_option(version=terrain_version, message='%(version)s')
@click.pass_context
def slice(ctx, input, output, minimum, maximum, keep_data, zeros, njobs, verbose):
    """Extract areas from a raster by a data range.

    Setting the --keep-data option will return the data values.
    The default is to return a raster of ones and zeros.

    \b
    Example:
    rio range diff.tif extracted.tif --minumum -2.0 --maximum 2.0

    """

    t0 = clock()

    with rasterio.Env():

        with rasterio.open(input) as src:

            profile = src.profile
            affine = src.transform
            step = (affine[0], affine[4])

            if keep_data:
                dtype = profile['dtype']
                nodata = profile['nodata']
                profile.update(count=1, compress='lzw')
            else:
                dtype = 'int32'
                nodata = np.iinfo(np.int32).min
                profile.update(dtype=rasterio.int32, nodata=nodata, count=1,
                               compress='lzw')

            if zeros:
                false_val = 0
            else:
                false_val = nodata

            with rasterio.open(output, 'w', **profile) as dst:
                if njobs < 1:
                    click.echo(msg.INMEMORY)
                    data = src.read(1)
                    result = _slice(data, minimum, maximum, keep_data, false_val)
                    dst.write(result.astype(dtype), 1)
                elif njobs == 1:
                    click.echo(msg.SEQUENTIAL)
                    for (ij, window) in src.block_windows():
                        data = src.read(1, window=window)
                        result = _slice(data, minimum, maximum, keep_data, false_val)
                        dst.write(result.astype(dtype), 1, window=window)
                else:
                    click.echo(msg.CONCURRENT)

                    def jobs():
                        for (ij, window) in src.block_windows():
                            data = src.read(1, window=window)
                            yield data, window

                    with concurrent.futures.ThreadPoolExecutor(max_workers=njobs) as executor:

                        future_to_window = {
                            executor.submit(
                                _slice,
                                data,
                                minimum,
                                maximum,
                                keep_data,
                                false_val): (window)
                            for (data, window) in jobs()}

                        for future in concurrent.futures.as_completed(future_to_window):
                            window = future_to_window[future]
                            result = future.result()
                            dst.write(result.astype(dtype), 1, window=window)

    click.echo('Wrote extracted raster to {}'.format(output))

    t1 = clock()
    click.echo('Finished in : {}'.format(msg.printtime(t0, t1)))



import numpy as np


def mad(arr, size=(3, 3)):
    """Calculates the median absolute deviation (MAD) for an array

    """
    from scipy.ndimage.filters import median_filter

    medians = median_filter(arr, size=size)
    deviations = np.absolute(arr - medians)
    mads = median_filter(deviations, size=size)

    return mads


def std(arr, size=(3, 3)):
    """Calculates the standard deviation for a neighborhood

    """
    from scipy.signal import convolve2d

    k = np.ones(size)

    arr_f64 = arr.astype(np.float64)

    ones = np.ones(arr.shape)
    c1 = convolve2d(arr_f64, k, mode="same")
    c2 = convolve2d(arr_f64*arr_f64, k, mode="same")
    ns = convolve2d(ones, k, mode="same")

    result = np.sqrt((c2 - c1**2 / ns) / ns)

    return result.astype(np.float32)


def std_ndimage(arr, size=(3, 3)):
    """Calculates the standard deviation for a neighborhood.

    """
    from scipy.ndimage.filters import convolve

    k = np.ones(size)

    arr_f64 = arr.astype(np.float64)

    c1 = convolve(arr_f64, k, mode='constant', cval=np.nan, origin=0)
    c2 = convolve(arr_f64*arr_f64, k, mode='constant', cval=np.nan, origin=0)

    vals = c2 - c1*c1
    result = (vals**.5)

    return result



from math import floor, ceil

import numpy as np
from rasterio.windows import Window
from rasterio.transform import rowcol, xy, from_bounds


def tile_dim(dim, tile_size, min_size=None):
    """Chunk a range using a minimum chunk size

    """
    coords = np.arange(0, dim, tile_size)
    if min_size and (dim % tile_size < min_size):
        coords = coords[:-1]

    return coords


def tile_dims(shape, tile_shape, min_size=None):
    """Chunk a 2D array using a minimum chunk size

    """
    xx = tile_dim(shape[0], tile_shape[0], min_size=min_size)
    yy = tile_dim(shape[1], tile_shape[1], min_size=min_size)

    return xx, yy


def chunk_dim(dim, chunk_size, min_size=None):
    nchunks = floor(dim/chunk_size)
    remainder = dim % chunk_size

    if nchunks == 0:
        chunks = [remainder]
    elif min_size and (nchunks > 1) and (remainder < min_size):
        chunks = [chunk_size]*(nchunks-1)
        chunks.append(chunk_size+remainder)
    else:
        chunks = [chunk_size]*nchunks
        chunks.append(remainder)

    return np.array(chunks)


def chunk_dims(shape, chunk_shape, min_size=None):

    hh = chunk_dim(shape[0], chunk_shape[0], min_size=min_size)
    ww = chunk_dim(shape[1], chunk_shape[1], min_size=min_size)

    return hh, ww


def block_count(shape, block_shapes, band=1):
    """Determine the number of blocks in a band

    Parameters:
        shape (tuple) : tuple containing raster height and width in cells
        block_shapes (tuple) : block shapes for a rasterio read source
        band (int) : raster band to count on

    Returns:
        result (int) : number of blocks in the raster

    """
    block_shape = block_shapes[band-1]

    blocks_wide = ceil(shape[1]/block_shape[1])
    blocks_high = ceil(shape[0]/block_shape[0])

    return blocks_high*blocks_wide


def subsample(blocks, probability=1.0):
    """Subsample an iterable at a given probability

    Parameters:
        blocks (iterable) : an iterable of rasterio windows
        probability (float) : fraction of blocks to sample

    Yields:
        block (window) : yield a rasterio window if sampled

    """
    import random

    for block in blocks:
        if random.random() < probability:
            yield block


def expand_window(window, src_shape, margin=10):
    """Expand a window by a margin

    Parameters:
        window (Window)
        src_shape (tuple)
        margin (int)

    Returns:
        result (Window)

    """
    cols, rows = window.toslices()

    row_start = rows.start - margin
    row_stop = rows.stop + margin
    col_start = cols.start - margin
    col_stop = cols.stop + margin

    row_start = 0 if (row_start < 0) else row_start
    col_start = 0 if col_start < 0 else col_start
    row_stop = src_shape[1] if row_stop > src_shape[1] else row_stop
    col_stop = src_shape[0] if col_stop > src_shape[0] else col_stop

    result = Window.from_slices(slice(col_start, col_stop),
                                slice(row_start, row_stop))

    return result


def bounds_window(bounds, affine, constrain=True):
    """Create a full cover rasterio-style window

    Parameters:
        bounds (tuple)
        affine (Affine)

    Returns:
        row_slice (tuple)
        col_slice (tuple)

    """
    w, s, e, n = bounds
    row_start, col_start = rowcol(affine, w, n)
    row_stop, col_stop = rowcol(affine, e, s, op=ceil)

    if constrain:
        if row_start < 0:
            row_start = 0
        if col_start < 0:
            col_start = 0

    return (row_start, row_stop), (col_start, col_stop)


def slices_to_window(rows, cols):

    return Window.from_slices(rows, cols)


def window_bounds(window, affine, offset='center'):
    """Create bounds coordinates from a rasterio window

    Parameters:
        window (Window)
        affine (Affine)
        offset (str)

    Returns:
        bounds (tuple) : coordinate bounds (w, s, e, n)

    """
    (row_start, col_start), (row_stop, col_stop) = window
    w, s = xy(affine, row_stop, col_start, offset=offset)
    e, n = xy(affine, row_start, col_stop, offset=offset)
    bounds = (w, s, e, n)

    return bounds


def intersect_bounds(bbox0, bbox1):
    """Get the intersection in w s e n

    Parameters:
        bbox0 (tuple)
        bbox1 (tuple)

    Returns:
        bounds (tuple) : coordinate bounds (w, s, e, n)

    """
    w = max(bbox0[0], bbox1[0])
    s = max(bbox0[1], bbox1[1])
    e = min(bbox0[2], bbox1[2])
    n = min(bbox0[3], bbox1[3])
    bounds = (w, s, e, n)

    return bounds


def margins(window0, window1):
    """Size of collar between a pair of windows

    Here, window0 is a read window and window1 is a write window
    """
    cols0, rows0 = window0.toslices()
    cols1, rows1 = window1.toslices()

    left = cols1.start - cols0.start
    upper = rows1.start - rows0.start
    right = cols0.stop - cols1.stop
    lower = rows0.stop - rows1.stop

    return (int(left), int(upper), int(right), int(lower))


def trim(arr, margins):
    """Trim a 2D array by a set of margins

    """
    (left, upper, right, lower) = margins

    if (right == 0) and (lower == 0):
        result = arr[left:, upper:]
    elif right == 0:
        result = arr[left:, upper: -lower]
    elif lower == 0:
        result = arr[left: -right, upper:]
    else:
        result = arr[left: -right, upper: -lower]

    return result


def tile_grid(ncols, nrows, blockxsize, blockysize,
              col_offset=0, row_offset=0, overlap=0):
    """Return a generator containing read and write windows with a specified
    dimensions and overlap

    mgrid returns not as expected so used broadcast_arrays instead
    base_rows, base_cols = np.mgrid[0:h:blockysize, 0:w:blockxsize]

    Parameters:
        ncols (int) : raster width in columns
        nrows (int) : raster height in rows
        blockxsize (int) : block width in rows
        blockysize (int) : block height in rows
        col_offset (int) : columns to offset the grid
        row_offset (int) : rows to offset the grid
        overlap (int) : overlap between windows

    Yields:
        window (Window) : tiled windows over a region

    """
    rows, cols = tile_dims((ncols, nrows), (blockxsize, blockysize))
    base_cols, base_rows = np.broadcast_arrays(rows, cols.reshape(cols.shape[0], -1))

    ul_cols = base_cols - overlap
    ul_rows = base_rows - overlap
    lr_cols = base_cols + blockxsize + overlap
    lr_rows = base_rows + blockysize + overlap

    ul_cols[ul_cols < 0] = 0
    ul_rows[ul_rows < 0] = 0
    lr_cols[lr_cols > ncols] = ncols
    lr_rows[lr_rows > nrows] = nrows

    for (ul_row, ul_col, lr_row, lr_col) in zip(ul_rows.ravel()+row_offset,
                                                ul_cols.ravel()+col_offset,
                                                lr_rows.ravel()+row_offset,
                                                lr_cols.ravel()+col_offset):

        yield Window.from_slices(slice(ul_row, lr_row), slice(ul_col, lr_col))


def tile_grid_intersection(src0, src1, blockxsize=None, blockysize=None):
    """Generate tiled windows for the intersection between two grids.

    Given two rasters having different dimensions calculate read-window generators for each
    and a write-window generator for the intersecion.

    Parameters:
        src0 : rasterio read source
        src1 : rasterio read source
        blockxsize (int) : write-window width
        blockysize (int) : write-window height

    Returns:
        src0_blocks : read windows for src0
        src1_blocks : read windows for src1
        write_blocks : write windows for the intersection
        affine (Affine) : write raster Affine
        ncols (int) : write raster width in columns
        nrows (int) : write raster height in rows

    """
    # transform to coordinates
    bbox0 = window_bounds(((0, 0), src0.shape),
                          src0.transform, offset='ul')

    bbox1 = window_bounds(((0, 0), src1.shape),
                          src1.transform, offset='ul')

    bounds = intersect_bounds(bbox0, bbox1)

    (row_start0, row_stop0), (col_start0, col_stop0) = bounds_window(bounds, src0.transform)
    (row_start1, row_stop1), (col_start1, col_stop1) = bounds_window(bounds, src1.transform)

    ncols = col_stop0 - col_start0
    nrows = row_stop0 - row_start0
    affine = from_bounds(bounds[0], bounds[1], bounds[2], bounds[3], ncols, nrows)

    if blockxsize is None:
        blockxsize = ncols
    if blockysize is None:
        blockysize = nrows

    windows0 = tile_grid(ncols, nrows, blockxsize, blockysize,
                         col_offset=col_start0, row_offset=row_start0)
    windows1 = tile_grid(ncols, nrows, blockxsize, blockysize,
                         col_offset=col_start1, row_offset=row_start1)
    write_windows = tile_grid(ncols, nrows, blockxsize, blockysize)

    return (windows0, windows1, write_windows, affine, nrows, ncols)


def is_raster_congruent(src0, src1, band=1):
    """Tests two rasters for coincident bounds.

    Parameters:
        src0 : rasterio read source
        src1 : rasterio read source

    Returns:
        result (bool) : True if the rasters are coincident

    """
    window0 = ((0, 0), src0.shape)
    window1 = ((0, 0), src1.shape)

    # get the affine
    affine0 = src0.transform
    affine1 = src1.transform

    return (window0 == window1) & (affine0 == affine1)


def is_raster_aligned(src0, src1):
    """Check two rasters for cell alignment.

    Parameters:
        src0 : rasterio read source
        src1 : rasterio read source

    Returns:
        result (bool) : True if the raster source cells align

    """
    affine0 = np.array(src0.transform)
    affine1 = np.array(src1.transform)

    step0 = affine0[0::3]
    step1 = affine1[0::3]
    rot0 = affine0[1::3]
    rot1 = affine1[1::3]
    ul0 = affine0[2::3]
    ul1 = affine1[2::3]

    return np.array_equal(step0, step1) & np.array_equal(rot0, rot1) & ((ul0 - ul1) % step0 == 0).all()



from time import clock
import warnings

import concurrent.futures
import multiprocessing

import click
import numpy as np
import rasterio

import rio_terrain as rt
import rio_terrain.tools.messages as msg
from rio_terrain import __version__ as terrain_version


@click.command()
@click.argument('input', nargs=1, type=click.Path(exists=True))
@click.argument('output', nargs=1, type=click.Path())
@click.option('--neighbors', type=click.Choice(['4', '8']), default='4',
              help='Specifies the number of neighboring cells to use.')
@click.option('-u', '--units', type=click.Choice(['grade','degrees']), default='grade',
              help='Specifies the units of slope.')
@click.option('-j', '--njobs', type=int, default=multiprocessing.cpu_count(),
              help='Number of concurrent jobs to run.')
@click.option('-v', '--verbose', is_flag=True, help='Enables verbose mode.')
@click.version_option(version=terrain_version, message='%(version)s')
@click.pass_context
def slope(ctx, input, output, neighbors, units, njobs, verbose):
    """Calculates slope of a height raster.

    \b
    Example:
    rio slope elevation.tif slope.tif

    """

    t0 = clock()

    with rasterio.Env():

        with rasterio.open(input) as src:
            profile = src.profile
            affine = src.transform
            step = (affine[0], affine[4])
            profile.update(dtype=rasterio.float32, count=1, compress='lzw')

            if njobs >= 1:
                blockshape = (list(src.block_shapes))[0]
                if (blockshape[0] == 1) or (blockshape[1] == 1):
                    warnings.warn((msg.STRIPED).format(blockshape))
                read_windows = rt.tile_grid(src.width, src.height, blockshape[0], blockshape[1], overlap=2)
                write_windows = rt.tile_grid(src.width, src.height, blockshape[0], blockshape[1], overlap=0)

            with rasterio.open(output, 'w', **profile) as dst:

                if njobs < 1:
                    click.echo(msg.INMEMORY)

                    data = src.read(1)
                    data[data <= src.nodata+1] = np.nan
                    result = rt.slope(data, step=step, units=units, neighbors=int(neighbors))
                    dst.write(result, 1)
                elif njobs == 1:
                    click.echo(msg.SEQUENTIAL)

                    # for (_, read_window, write_window, olap) in rt.tiled_windows(src.height, src.width, block_width=128, block_height=128, overlap=2):
                    for (read_window, write_window) in zip(read_windows, write_windows):
                        data = src.read(1, window=read_window)
                        data[data <= src.nodata+1] = np.nan
                        arr = rt.slope(data, step=step, units=units, neighbors=int(neighbors))
                        result = rt.trim(arr, rt.margins(read_window, write_window))
                        dst.write(result, 1, window=write_window)
                else:
                    click.echo(msg.CONCURRENT)

                    def jobs():
                        for (read_window, write_window) in zip(read_windows, write_windows):
                            data = src.read(1, window=read_window)
                            data[data <= src.nodata+1] = np.nan
                            yield data, read_window, write_window

                    with concurrent.futures.ThreadPoolExecutor(max_workers=njobs) as executor:

                        future_to_window = {
                            executor.submit(
                                rt.slope,
                                data,
                                step=step,
                                units=units,
                                neighbors=int(neighbors)): (read_window, write_window)
                            for (data, read_window, write_window) in jobs()}

                        for future in concurrent.futures.as_completed(future_to_window):
                            read_window, write_window = future_to_window[future]
                            arr = future.result()
                            result = rt.trim(arr, rt.margins(read_window, write_window))
                            dst.write(result, 1, window=write_window)

    t1 = clock()
    click.echo('Finished in : {}'.format(msg.printtime(t0, t1)))



import concurrent.futures

import numpy as np


def minmax(src, windows, njobs):
    """Calculates the minimum and maximum values in a rasterio source.

    Args:
        src : rasterio source
        windows : iterable of read and write windows
        njobs (integer) : number of processing jobs

    Returns:
        src_min (float) : minimum value
        src_max (float) : maximum value

    ArcGIS min = 77.278923034668
    ArcGIS max = 218.81454467773
    """

    def _minmax(arr):
        mask = np.isfinite(arr[:])
        if np.count_nonzero(mask) > 0:
            arr_min = np.nanmin(arr[mask])
            arr_max = np.nanmax(arr[mask])
        else:
            arr_min = None
            arr_max = None

        return arr_min, arr_max

    src_min = None
    src_max = None

    if njobs < 1:
        data = src.read(1)
        data[data <= src.nodata+1] = np.nan
        src_min, src_max = _minmax(data)
        return src_min, src_max
    else:
        def jobs():
            for ij, window in windows:
                data = src.read(1, window=window)
                data[data <= src.nodata+1] = np.nan
                yield data, window

        with concurrent.futures.ThreadPoolExecutor(max_workers=njobs) as executor:

            future_to_window = {
                executor.submit(
                    _minmax,
                    data): (window)
                for data, window in jobs()}

            for future in concurrent.futures.as_completed(future_to_window):
                # window = future_to_window[future]
                window_min, window_max = future.result()

                if not src_min and not src_max:
                    if window_min and window_max:
                        src_min = window_min
                        src_max = window_max
                elif window_min and window_max:
                    if window_min < src_min:
                        src_min = window_min
                    if window_max > src_max:
                        src_max = window_max

        return src_min, src_max


def mean(src, windows, njobs):
    """Calculates the mean of a rasterio source

    Args:
        src : rasterio source
        windows : iterable of read and write windows
        njobs (integer) : number of processing jobs

    Returns:
        mean (float) : mean value

    mean =  140.043719088
    ArcGIS = 140.04371922353
    """

    def _accumulate(arr):
        mask = np.isfinite(arr[:])
        sum_ = np.sum(arr[mask])
        count_ = np.count_nonzero(mask)

        return sum_, count_

    if njobs < 1:
        data = src.read(1)
        data[data <= src.nodata+1] = np.nan
        vals = data[np.isfinite(data[:])]
        mean = np.nanmean(vals)
    else:

        def jobs():
            for ij, window in windows:
                data = src.read(1, window=window)
                data[data <= src.nodata+1] = np.nan
                yield data, window

        src_sum = 0.0
        src_count = 0.0

        with concurrent.futures.ThreadPoolExecutor(max_workers=njobs) as executor:

            future_to_window = {
                executor.submit(
                    _accumulate,
                    data): (window)
                for data, window in jobs()}

            for future in concurrent.futures.as_completed(future_to_window):
                # window = future_to_window[future]
                window_sum, window_count = future.result()
                src_sum += window_sum
                src_count += window_count

            mean = src_sum / src_count

    return mean


def stddev(src, mean, windows, njobs):
    """Calculates the standard deviation of a rasterio source

    Args:
        src : rasterio source
        mean : mean value
        windows : iterable of read and write windows
        njobs (integer) : number of processing jobs

    Returns:
        stddev (float) : standard deviation

    stddev =  23.5554506735
    ArcGIS = 23.555450665488
    """
    def _accumulate(arr):
        mask = np.isfinite(arr[:])
        sum_ = np.sum(np.square(arr[mask]-mean))
        count_ = np.count_nonzero(mask)

        return sum_, count_

    if njobs < 1:
        data = src.read(1)
        data[data <= src.nodata+1] = np.nan
        vals = data[np.isfinite(data[:])]
        stddev = np.nanstd(vals)
    else:

        def jobs():
            for ij, window in windows:
                data = src.read(1, window=window)
                data[data <= src.nodata+1] = np.nan
                yield data, window

        src_dev_sum = 0.0
        src_dev_count = 0.0

        with concurrent.futures.ThreadPoolExecutor(max_workers=njobs) as executor:

            future_to_window = {
                executor.submit(
                    _accumulate,
                    data): (window)
                for data, window in jobs()}

            for future in concurrent.futures.as_completed(future_to_window):
                # window = future_to_window[future]
                window_dev_sum, window_dev_count = future.result()
                src_dev_sum += window_dev_sum
                src_dev_count += window_dev_count

            stddev = np.sqrt(src_dev_sum / (src_dev_count-1))

    return stddev



from time import clock
import warnings
import concurrent.futures
import multiprocessing

import click
import numpy as np
import rasterio

import rio_terrain as rt
import rio_terrain.tools.messages as msg
from rio_terrain import __version__ as terrain_version


def _extract(data, categorical, category):
    if category is None:
        _category = [1]
    else:
        _category = list(category)
    mask = np.isin(categorical, _category)
    result = data*mask
    return result


@click.command()
@click.argument('input', nargs=1, type=click.Path(exists=True))
@click.argument('categorical', nargs=1, type=click.Path(exists=True))
@click.argument('output', nargs=1, type=click.Path())
@click.option('-c', '--category', multiple=True, type=int,
              help='Category to extract.')
@click.option('-j', '--njobs', type=int, default=multiprocessing.cpu_count(),
              help='Number of concurrent jobs to run')
@click.option('-v', '--verbose', is_flag=True, help='Enables verbose mode.')
@click.version_option(version=terrain_version, message='%(version)s')
@click.pass_context
def extract(ctx, input, categorical, output, category, njobs, verbose):
    """Extract areas from a raster by category.

    The categorical raster may be the input raster or another raster.

    \b
    Example:
    rio extract diff.tif categorical.tif extract.tif -c 1 -c 3

    """

    t0 = clock()

    with rasterio.Env():

        with rasterio.open(input) as src, \
                rasterio.open(categorical) as cat:

            profile = src.profile
            affine = src.transform
            step = (affine[0], affine[4])

            if njobs >= 1:
                block_shape = (src.block_shapes)[0]
                blockxsize = block_shape[1]
                blockysize = block_shape[0]
            else:
                blockxsize = None
                blockysize = None

            tiles = rt.tile_grid_intersection(src, cat, blockxsize=blockxsize, blockysize=blockysize)
            windows0, windows1, write_windows, affine, nrows, ncols = tiles

            profile.update(count=1, compress='lzw', bigtiff='yes',
                           height=nrows, width=ncols, transform=affine)

            with rasterio.open(output, 'w', **profile) as dst:
                if njobs < 1:
                    click.echo(msg.INMEMORY)
                    data = src.read(1, window=next(windows0))
                    mask = cat.read(1, window=next(windows1))
                    result = _extract(data, mask, category)
                    dst.write(result, 1, window=next(write_windows))
                elif njobs == 1:
                    click.echo(msg.SEQUENTIAL)
                    for (window0, window1, write_window) in zip(windows0, windows1, write_windows):
                        data = src.read(1, window=window0)
                        mask = cat.read(1, window=window1)
                        result = _extract(data, mask, category)
                        dst.write(result, 1, window=write_window)
                else:
                    click.echo(msg.CONCURRENT)

                    def jobs():
                        for (window0, window1, write_window) in zip(windows0, windows1, write_windows):
                            data = src.read(1, window=window0)
                            mask = cat.read(1, window=window1)
                            yield data, mask, window0, window1, write_window

                    with concurrent.futures.ThreadPoolExecutor(max_workers=njobs) as executor:

                        future_to_window = {
                            executor.submit(
                                _extract,
                                data,
                                mask,
                                category): (window0, window1, write_window)
                            for (data, mask, window0, window1, write_window) in jobs()}

                        for future in concurrent.futures.as_completed(future_to_window):
                            window0, window1, write_window = future_to_window[future]
                            result = future.result()
                            dst.write(result, 1, window=write_window)

    click.echo('Wrote extracted raster to {}'.format(output))

    t1 = clock()
    click.echo('Finished in : {}'.format(msg.printtime(t0, t1)))



from time import clock
import warnings
import concurrent.futures
import multiprocessing

import click
import numpy as np
import rasterio

import rio_terrain as rt
import rio_terrain.tools.messages as msg
from rio_terrain import __version__ as terrain_version


def _propagate(data0, data1, instrumental0, instrumental1):
    if instrumental0:
        data0[data0 < instrumental0] = instrumental0
    if instrumental1:
        data1[data1 < instrumental1] = instrumental1

    result = np.sqrt(np.square(data1) + np.square(data0))

    return result


@click.command()
@click.argument('uncertainty0', nargs=1, type=click.Path(exists=True))
@click.argument('uncertainty1', nargs=1, type=click.Path(exists=True))
@click.argument('output', nargs=1, type=click.Path())
@click.option('--instrumental0', nargs=1, default=None, type=float,
              help='Instrumental or minimum uncertainty for the first raster.')
@click.option('--instrumental1', nargs=1, default=None, type=float,
              help='Instrumental or minimum uncertainty for the second raster.')
@click.option('-j', '--njobs', type=int, default=multiprocessing.cpu_count(),
              help='Number of concurrent jobs to run.')
@click.option('-v', '--verbose', is_flag=True, help='Enables verbose mode.')
@click.version_option(version=terrain_version, message='%(version)s')
@click.pass_context
def uncertainty(ctx, uncertainty0, uncertainty1, output, instrumental0, instrumental1, njobs, verbose):
    """Calculates a minimum level of detection raster.

    \b
    Example:
    rio uncertainty roughness_t0.tif roughness_t1.tif uncertainty.tif

    """

    t0 = clock()

    with rasterio.Env():

        with rasterio.open(uncertainty0) as src0, rasterio.open(uncertainty1) as src1:
            profile = src0.profile
            affine = src0.transform
            step = (affine[0], affine[4])

            if njobs >= 1:
                block_shape = (src0.block_shapes)[0]
                blockxsize = block_shape[1]
                blockysize = block_shape[0]
            else:
                blockxsize = None
                blockysize = None

            tiles = rt.tile_grid_intersection(src0, src1, blockxsize=blockxsize, blockysize=blockysize)
            windows0, windows1, write_windows, affine, nrows, ncols = tiles

            profile.update(dtype=rasterio.float32, count=1,
                           height=nrows, width=ncols, transform=affine,
                           compress='lzw', bigtiff='yes')

            with rasterio.open(output, 'w', **profile) as dst:
                if njobs < 1:
                    click.echo(msg.INMEMORY)
                    data0 = src0.read(1)
                    data1 = src1.read(1)
                    result = _propagate(data0, data1, instrumental0, instrumental1)
                    dst.write(result.astype(np.float32), 1)
                elif njobs == 1:
                    click.echo(msg.SEQUENTIAL)
                    for (window0, window1, write_window) in zip(windows0, windows1, write_windows):
                        data0 = src0.read(1, window=window0)
                        data1 = src1.read(1, window=window1)
                        result = _propagate(data0, data1, instrumental0, instrumental1)
                        dst.write(result.astype(np.float32), 1, window=write_window)
                else:
                    click.echo(msg.CONCURRENT)

                    def jobs():
                        for (window0, window1, write_window) in zip(windows0, windows1, write_windows):
                            data0 = src0.read(1, window=window0)
                            data1 = src1.read(1, window=window1)
                            yield data0, data1, window0, window1, write_window

                    with concurrent.futures.ThreadPoolExecutor(max_workers=njobs) as executor:

                        future_to_window = {
                            executor.submit(
                                _propagate,
                                data0,
                                data1,
                                instrumental0,
                                instrumental1): (write_window)
                            for (data0, data1, window0, window1, write_window) in jobs()}

                        for future in concurrent.futures.as_completed(future_to_window):
                            write_window = future_to_window[future]
                            result = future.result()
                            dst.write(result.astype(np.float32), 1, window=write_window)

    click.echo('Wrote uncertainty raster to {}'.format(output))

    t1 = clock()
    click.echo('Finished in : {}'.format(msg.printtime(t0, t1)))



from pkg_resources import iter_entry_points

from click.testing import CliRunner

import rasterio
from rasterio.rio.main import main_group

import rio_terrain


def test_version():
    runner = CliRunner()
    result = runner.invoke(main_group, ['aspect', '--version'])
    assert result.exit_code == 0
    assert rio_terrain.__version__ in result.output


def test_all_registered():
    # This test makes sure that all of the subcommands defined in the
    # rasterio.rio_commands entry-point are actually registered to the main
    # cli group.
    for ep in iter_entry_points('rasterio.rio_commands'):
        assert ep.name in main_group.commands


def test_aspect():
    runner = CliRunner()
    result = runner.invoke(main_group, ['aspect', '--help'])
    assert result.exit_code == 0


def test_curvature():
    runner = CliRunner()
    result = runner.invoke(main_group, ['curvature', '--help'])
    assert result.exit_code == 0


def test_difference():
    runner = CliRunner()
    result = runner.invoke(main_group, ['difference', '--help'])
    assert result.exit_code == 0


def test_extract():
    runner = CliRunner()
    result = runner.invoke(main_group, ['extract', '--help'])
    assert result.exit_code == 0


def test_mad():
    runner = CliRunner()
    result = runner.invoke(main_group, ['mad', '--help'])
    assert result.exit_code == 0


def test_quantiles():
    runner = CliRunner()
    result = runner.invoke(main_group, ['quantiles', '--help'])
    assert result.exit_code == 0


def test_slice():
    runner = CliRunner()
    result = runner.invoke(main_group, ['slice', '--help'])
    assert result.exit_code == 0


def test_slope():
    runner = CliRunner()
    result = runner.invoke(main_group, ['slope', '--help'])
    assert result.exit_code == 0


def test_std():
    runner = CliRunner()
    result = runner.invoke(main_group, ['std', '--help'])
    assert result.exit_code == 0


def test_threshold():
    runner = CliRunner()
    result = runner.invoke(main_group, ['threshold', '--help'])
    assert result.exit_code == 0


def test_uncertainty():
    runner = CliRunner()
    result = runner.invoke(main_group, ['uncertainty', '--help'])
    assert result.exit_code == 0



from setuptools import setup, find_packages


# Parse the version from the shapely module
for line in open('rio_terrain/__init__.py', 'r'):
    if line.find("__version__") >= 0:
        version = line.split("=")[1].strip()
        version = version.strip('"')
        version = version.strip("'")
        continue

with open('VERSION.txt', 'w') as fp:
    fp.write(version)

setup(name='rio-terrain',
      version=version,
      author='Michael Rahnis',
      author_email='michael.rahnis@fandm.edu',
      description='Rasterio CLI plugin to perform common operations on height rasters.',
      url='http://github.com/mrahnis/rio-terrain',
      license='BSD',
      packages=find_packages(exclude=['examples']),
      include_package_data=True,
      install_requires=[
        'numpy',
        'scipy',
        'matplotlib',
        'rasterio',
        'click',
        'crick',
      ],
      entry_points='''
        [rasterio.rio_commands]
        aspect=rio_terrain.cli.aspect:aspect
        curvature=rio_terrain.cli.curvature:curvature
        difference=rio_terrain.cli.difference:difference
        extract=rio_terrain.cli.extract:extract
        mad=rio_terrain.cli.mad:mad
        quantiles=rio_terrain.cli.quantiles:quantiles
        slice=rio_terrain.cli.slice:slice
        slope=rio_terrain.cli.slope:slope
        std=rio_terrain.cli.std:std
        threshold=rio_terrain.cli.threshold:threshold
        uncertainty=rio_terrain.cli.uncertainty:uncertainty
      ''',
      keywords='LiDAR, surveying, elevation, raster, DEM',
      classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: GIS'
      ])



import logging

from rio_terrain.core.terrain import *
from rio_terrain.core.statistics import *
from rio_terrain.core.windowing import *

__author__ = "Mike Rahnis"
__version__ = '0.0.11'


logger = logging.getLogger(__name__)



import sys

# Information
INMEMORY = "Processing raster in-memory..." 
SEQUENTIAL = "Processing raster blocks sequentially..."
CONCURRENT = "Processing raster blocks concurrently..."

# Warning
STRIPED = 'Blocks are lines with shape {}. Rewrite the data blocks for faster out-of-core processing.'


def printtime(t0, t1):
    '''Return the elapsed time between t0 and t1 in h:m:s formatted string

    '''
    m, s = divmod(t1 - t0, 60)
    h, m = divmod(m, 60)
    fmt = '%d:%02d:%02d' % (h, m, s)

    return fmt


def progress(value, endvalue, bar_length=20, msg=None):
    """Display and update a progress bar

    """
    done_char = '#'
    todo_char = ' '

    percent = float(value) / endvalue
    progress = done_char * int(round(percent * bar_length))
    spaces = todo_char * (bar_length - len(progress))

    message = "\rPercent: [{0}] {1}%    {2}".format(progress + spaces, int(round(percent * 100)), msg)
    sys.stdout.write(message)
    if value == endvalue-1:
        sys.stdout.write('\n')
    sys.stdout.flush()



import warnings
from math import pi

import numpy as np
from scipy import ndimage


def _ring_gradient(arr, step=(1, 1)):
    """Convolve an array using a 3x3 ring-shaped kernel

    Parameters:
        arr (ndarray): 2D numpy array
        step (tuple): tuple of raster cell width and height

    Returns:
        dz_dx, dz_dy (ndarrays): x and y gradient components

    """
    origin = (0, 0)

    k_X = np.array([[-1, 0, 1],
                    [-2, 0, 2],
                    [-1, 0, 1]])
    k_Y = np.array([[-1, -2, -1],
                    [0, 0, 0],
                    [1, 2, 1]])

    dz_dx = ndimage.convolve(arr, k_X, origin=origin) / (8 * step[0])
    dz_dy = ndimage.convolve(arr, k_Y, origin=origin) / (8 * step[1])

    return dz_dx, dz_dy


def slope(arr, step=(1, 1), units='grade', neighbors=4):
    """Calculates slope.

    Parameters:
        arr (ndarray): 2D numpy array
        step (tuple): tuple of raster cell width and height
        units (str, optional): choice of grade or degrees
        neighbors (int, optional): use four or eight neighbors in calculation

    Returns:
        slope (ndarray): 2D numpy array representing slope

    """
    if neighbors == 4:
        dz_dx, dz_dy = np.gradient(arr, step[0])
    else:
        dz_dx, dz_dy = _ring_gradient(arr, step)

    m = np.sqrt(dz_dx**2 + dz_dy**2)
    if units == 'grade':
        slope = m
    elif units == 'degrees':
        slope = (180/pi) * np.arctan(m)

    return slope


def aspect(arr, step=(1, 1), pcs='compass', neighbors=4):
    """Calculates aspect.

    Parameters:
        arr (ndarray): 2D numpy array
        step (tuple): tuple of raster cell width and height
        north (str, optional): choice of polar coordinate systeM

    Returns:
        aspect (ndarray): 2D numpy array representing slope aspect

    """
    if neighbors == 4:
        dz_dx, dz_dy = np.gradient(arr, step[0])
    else:
        dz_dx, dz_dy = _ring_gradient(arr, step)

    if pcs == 'compass':
        aspect = (180/pi)*np.arctan2(dz_dy, -dz_dx)
        aspect += 180
    elif pcs == 'cartesian':
        aspect = (180/pi)*np.arctan2(-dz_dy, dz_dx)
        aspect = 90.0 - aspect
        aspect[aspect < 0] += 360
    else:
        aspect = (180/pi)*np.arctan2(dz_dy, dz_dx)

    return aspect


def curvature(arr, step=(1, 1), neighbors=4):
    """Calculates curvature.

    Parameters:
        arr (ndarray): 2D numpy array
        step (tuple): tuple of raster cell width and height

    Returns:
        curvature (ndarray): 2D numpy array representing surface curvature

    """
    if neighbors == 4:
        dz_dx, dz_dy = np.gradient(arr, step[0])
    else:
        dz_dx, dz_dy = _ring_gradient(arr, step)

    m = np.sqrt(dz_dx**2 + dz_dy**2)
    dT_dx = np.divide(dz_dx, m)
    dT_dy = np.divide(dz_dy, m)
    d2T_dxx, _ = np.gradient(dT_dx, step[0])
    _, d2T_dyy = np.gradient(dT_dy, step[0])

    curvature = d2T_dxx + d2T_dyy

    return curvature



import numpy as np


def fix_nodata(arr, nodata):
    arr[arr <= nodata+1] = np.nan
    return arr


def is_all_nan(arr):
    if np.isnan(arr).all():
        return True
    else:
        return False


def nan_shape(shape):
    result = np.empty(shape)
    result[:] = np.nan
    return result
